/*******************************************************************************
 * Copyright (c) 2009 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.e4.languages.javascript.debug.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.model.Breakpoint;
import org.eclipse.debug.core.model.IDebugTarget;
import org.eclipse.debug.core.model.ILineBreakpoint;
import org.eclipse.e4.languages.javascript.debug.connect.JSONConstants;
import org.eclipse.e4.languages.javascript.jsdi.ScriptReference;
import org.eclipse.e4.languages.javascript.jsdi.event.BreakpointEvent;
import org.eclipse.e4.languages.javascript.jsdi.event.Event;
import org.eclipse.e4.languages.javascript.jsdi.event.EventSet;
import org.eclipse.e4.languages.javascript.jsdi.event.ScriptLoadEvent;
import org.eclipse.e4.languages.javascript.jsdi.request.EventRequest;
import org.eclipse.e4.languages.javascript.jsdi.request.ScriptLoadRequest;
import org.osgi.framework.Constants;

/**
 * Abstract representation of a JSDI breakpoint
 * 
 * @since 1.0
 */
public abstract class JSDIBreakpoint extends Breakpoint implements ILineBreakpoint, IJSDIEventListener {

	static final String OSGI_SYMBOLIC_NAME = "OSGi." + Constants.BUNDLE_SYMBOLICNAME; //$NON-NLS-1$

	/**
	 * Breakpoint attribute for the path of the script
	 */
	public static final String SCRIPT_PATH = "SCRIPT_PATH"; //$NON-NLS-1$

	/**
	 * The type name within the script
	 */
	public static final String TYPE_NAME = "TYPE_NAME"; //$NON-NLS-1$

	/**
	 * JSDT member handle
	 */
	public static final String HANDLE = "HANDLE"; //$NON-NLS-1$

	private HashSet targets = null;
	private HashMap requestspertarget = new HashMap(4);

	/**
	 * The id of the attribute for a method signature
	 */
	public static final String FUNCTION_SIGNAURE = "org.eclipse.e4.languages.javascript.methodsignature"; //$NON-NLS-1$

	/**
	 * The id of the attribute for the function name
	 */
	public static final String FUNCTION_NAME = "org.eclipse.e4.languages.javascript.functionname"; //$NON-NLS-1$

	/**
	 * Registered marker id for a JSDI line breakpoint
	 */
	public static final String JSDI_LINE_BREAKPOINT = "org.eclipse.e4.languages.javascript.line.breakpoint.marker"; //$NON-NLS-1$

	/**
	 * Registered marker id for a JSDI function breakpoint
	 */
	public static final String JSDI_FUNCTION_BREAKPOINT = "org.eclipse.e4.languages.javascript.function.breakpoint.marker"; //$NON-NLS-1$

	/**
	 * Add this breakpoint to the breakpoint manager, or sets it as unregistered.
	 */
	protected void register(boolean register) throws CoreException {
		DebugPlugin plugin = DebugPlugin.getDefault();
		if (plugin != null && register) {
			plugin.getBreakpointManager().addBreakpoint(this);
		} else {
			setRegistered(false);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.IBreakpoint#getModelIdentifier()
	 */
	public String getModelIdentifier() {
		return JSDIDebugModel.MODEL_ID;
	}

	/**
	 * Returns whether this breakpoint should be "skipped". Breakpoints are skipped if the breakpoint manager is disabled and the breakpoint is registered with the manager
	 * 
	 * @return whether this breakpoint should be skipped
	 */
	public boolean shouldSkipBreakpoint() throws CoreException {
		DebugPlugin plugin = DebugPlugin.getDefault();
		return plugin != null && isRegistered() && !plugin.getBreakpointManager().isEnabled();
	}

	/**
	 * Adds the given debug target to the listing of targets that this breakpoint cares about
	 * 
	 * @param target
	 * @throws CoreException
	 */
	public void addToTarget(JSDIDebugTarget target) throws CoreException {
		if (target.isTerminated() || shouldSkipBreakpoint()) {
			return;
		}
		// Add script load handler
		ScriptLoadRequest request = target.getEventRequestManager().createScriptLoadRequest();
		request.setEnabled(isEnabled());
		registerRequest(target, request);

		// Add to all loaded scripts
		String script = getScriptPath();
		if (script == null) {
			return;
		}
		script = new Path(script).lastSegment();
		List/* ScriptReference */scripts = target.allScriptsByName(script);
		boolean success = true;
		for (Iterator iter = scripts.iterator(); iter.hasNext();) {
			success &= createRequest(target, (ScriptReference) iter.next());
		}
		if (success) {
			if (this.targets == null) {
				this.targets = new HashSet();
			}
			this.targets.add(target);
		}
	}

	/**
	 * Creates a request for the given script in the given target
	 * 
	 * @param target
	 *            the target to register with
	 * @param script
	 *            the script we want to set the breakpoint in
	 * @return true if a new request was created false otherwise
	 * @throws CoreException
	 */
	protected abstract boolean createRequest(JSDIDebugTarget target, ScriptReference script) throws CoreException;

	/**
	 * Removes the given debug target from the listing of targets this breakpoint cares about
	 * 
	 * @param target
	 */
	public void removeFromTarget(JSDIDebugTarget target) {
		List requests = getRequests(target);
		EventRequest request = null;
		for (Iterator iter = requests.iterator(); iter.hasNext();) {
			try {
				request = (EventRequest) iter.next();
				if (target.isAvailable()) {
					target.getEventRequestManager().deleteEventRequest(request);
				}
			} finally {
				deregisterRequest(target, request);
			}
		}
		synchronized (this.requestspertarget) {
			this.requestspertarget.remove(target);
		}
		if (this.targets == null) {
			return;
		}
		this.targets.remove(target);
	}

	/**
	 * Recreate this breakpoint in all targets
	 * 
	 * @throws CoreException
	 */
	protected void recreateBreakpoint() throws CoreException {
		DebugPlugin plugin = DebugPlugin.getDefault();
		if (plugin != null) {
			IDebugTarget[] targets = plugin.getLaunchManager().getDebugTargets();
			for (int i = 0; i < targets.length; i++) {
				// TODO could also ask for the adapter
				if (targets[i].getModelIdentifier().equals(JSDIDebugModel.MODEL_ID)) {
					recreateBreakpointFor((JSDIDebugTarget) targets[i]);
				}
			}
		}
	}

	/**
	 * Recreates this breakpoint in the given debug target
	 * 
	 * @param target
	 * @throws CoreException
	 */
	void recreateBreakpointFor(JSDIDebugTarget target) throws CoreException {
		if (target.isAvailable() && target.getBreakpoints().contains(this)) {
			removeFromTarget(target);
			addToTarget(target);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.Breakpoint#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled) throws CoreException {
		recreateBreakpoint();
		super.setEnabled(enabled);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.ILineBreakpoint#getLineNumber()
	 */
	public int getLineNumber() throws CoreException {
		return ensureMarker().getAttribute(IMarker.LINE_NUMBER, -1);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.ILineBreakpoint#getCharStart()
	 */
	public int getCharStart() throws CoreException {
		return ensureMarker().getAttribute(IMarker.CHAR_START, -1);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.ILineBreakpoint#getCharEnd()
	 */
	public int getCharEnd() throws CoreException {
		return ensureMarker().getAttribute(IMarker.CHAR_END, -1);
	}

	/**
	 * Returns the name of the script as it was set when the breakpoint was created
	 * 
	 * @return the name of the script this breakpoint was created on
	 * @throws CoreException
	 */
	public String getScriptPath() throws CoreException {
		return ensureMarker().getAttribute(SCRIPT_PATH, null);
	}

	/**
	 * Returns the type name that the breakpoint is set within. When <code>null</code> is returned the breakpoint is set in the top level type i.e. the root source
	 * 
	 * @return the typeName
	 */
	public String getTypeName() throws CoreException {
		return ensureMarker().getAttribute(TYPE_NAME, null);
	}

	/**
	 * Registers the given request for this breakpoint for the given target
	 * 
	 * @param target
	 * @param request
	 */
	protected void registerRequest(JSDIDebugTarget target, EventRequest request) {
		ArrayList requests = getRequests(target);
		if (requests.isEmpty()) {
			synchronized (this.requestspertarget) {
				this.requestspertarget.put(target, requests);
			}
		}
		requests.add(request);
		target.addJSDIEventListener(this, request);
	}

	/**
	 * Returns any existing requests associated with the given target
	 * 
	 * @param target
	 * @return list of requests for the given target
	 */
	protected ArrayList getRequests(JSDIDebugTarget target) {
		ArrayList list = null;
		synchronized (this.requestspertarget) {
			list = (ArrayList) this.requestspertarget.get(target);
		}
		if (list == null) {
			list = new ArrayList(2);
		}
		return list;
	}

	/**
	 * Remove this breakpoint as an event listener
	 * 
	 * @param target
	 * @param request
	 */
	protected void deregisterRequest(JSDIDebugTarget target, EventRequest request) {
		target.removeJSDIEventListener(this, request);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.e4.languages.javascript.debug.model.IJSDIEventListener#handleEvent(org.eclipse.e4.languages.javascript.jsdi.event.Event, org.eclipse.e4.languages.javascript.debug.model.JSDIDebugTarget, boolean, org.eclipse.e4.languages.javascript.jsdi.event.EventSet)
	 */
	public boolean handleEvent(Event event, JSDIDebugTarget target, boolean suspendVote, EventSet eventSet) {
		// get the thread and suspend it
		if (event instanceof BreakpointEvent) {
			if (skipEvent((BreakpointEvent) event)) {
				return true;
			}
			JSDIThread thread = target.findThread(((BreakpointEvent) event).thread());
			if (thread != null) {
				return !thread.suspendForBreakpoint(this, suspendVote);
			}
		}
		if (event instanceof ScriptLoadEvent) {
			ScriptLoadEvent sevent = (ScriptLoadEvent) event;
			ScriptReference script = sevent.script();
			try {
				// TODO need something fancier in the future
				if (new Path(getScriptPath()).lastSegment().equals(script.sourcePath())) {
					createRequest(target, sevent.script());
				}
			} catch (CoreException ce) {
				// TODO log this
				ce.printStackTrace();
			}
		}
		return true;
	}

	/**
	 * If the breakpoint event should be skipped - i.e. if the thread should resume and ignore the event
	 * 
	 * @param event
	 * @return true if the event should be skipped, false otherwise
	 */
	boolean skipEvent(BreakpointEvent event) {
		try {
			Path root = new Path(getScriptPath());
			if (root.segmentCount() == 1) {
				// TODO hack to support external + virtual breakpoints
				// that have no path
				return false;
			}
			ScriptReference script = event.location().scriptReference();
			Map attribs = script.sourceProperties();
			if (attribs != null) {
				IPath path = new Path((String) attribs.get(OSGI_SYMBOLIC_NAME));
				path = path.append((String) attribs.get(JSONConstants.PATH));
				path = path.append((String) attribs.get(JSONConstants.NAME));
				return !pathsEqual(root, path);
			}
		} catch (CoreException ce) {
			// TODO log this
		}
		return true;
	}

	/**
	 * Returns if the type names for the breakpoint are equal or not. Two <code>null</code> type names are considered to be equal.
	 * 
	 * @param tname1
	 * @param tname2
	 * @return true if the type names are equal, false otherwise
	 */
	boolean typeNamesEqual(String tname1, String tname2) {
		if (tname1 == null && tname2 == null) {
			return true;
		}
		return tname1 != null && tname1.equals(tname2);
	}

	/**
	 * Custom comparison to avoid the leading separator issue from saying the paths are not equal
	 * 
	 * @param p1
	 * @param p2
	 * @return
	 */
	boolean pathsEqual(IPath p1, IPath p2) {
		if (p1.segmentCount() == p2.segmentCount()) {
			String[] segments = p1.segments();
			for (int i = 0; i < segments.length; i++) {
				if (!segments[i].equals(p2.segment(i))) {
					return false;
				}
			}
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.e4.languages.javascript.debug.model.IJSDIEventListener#eventSetComplete(org.eclipse.e4.languages.javascript.jsdi.event.Event, org.eclipse.e4.languages.javascript.debug.model.JSDIDebugTarget, boolean, org.eclipse.e4.languages.javascript.jsdi.event.EventSet)
	 */
	public void eventSetComplete(Event event, JSDIDebugTarget target, boolean suspend, EventSet eventSet) {
		if (event instanceof BreakpointEvent) {
			JSDIThread thread = target.findThread(((BreakpointEvent) event).thread());
			if (thread != null) {
				thread.suspendForBreakpointComplete(this, suspend, eventSet);
			}
		}
	}
}
